% test suite for builtin operations

-exports([main/0, e/2]).

e(X, Y) :- Y is X * 2.

ok({OK, NOTOK}, S) :- 
    OK2 is OK + 1, S := {OK2, NOTOK}.
notok({OK, NOTOK}, S) :- 
    NOTOK2 is NOTOK + 1, S := {OK, NOTOK2}.

r(T, E, E, S, S2) :- ok(S, S2).
r(T, R, E, S, S2) :- 
    R =\= E | 
    notok(S, S2), writeln(failed(T, R, E)).

eq(T, X, X, E, S, S2) :- r(T, yes, E, S, S2).
eq(T, X, Y, E, S, S2) :- X =\= Y | r(T, no, E, S, S2).
gt(T, X, Y, E, S, S2) :- X > Y | r(T, yes, E, S, S2).
gt(T, X, Y, E, S, S2) :- X =< Y | r(T, no, E, S, S2).
lt(T, X, Y, E, S, S2) :- X < Y | r(T, yes, E, S, S2).
lt(T, X, Y, E, S, S2) :- X >= Y | r(T, no, E, S, S2).
ge(T, X, Y, E, S, S2) :- X >= Y | r(T, yes, E, S, S2).
ge(T, X, Y, E, S, S2) :- X < Y | r(T, no, E, S, S2).
le(T, X, Y, E, S, S2) :- X =< Y | r(T, yes, E, S, S2).
le(T, X, Y, E, S, S2) :- X > Y | r(T, no, E, S, S2).
g_known(T, X, E, S, S2) :- known(X) | r(T, yes, E, S, S2).
g_known(T, X, E, S, S2) :- unknown(X) | r(T, no, E, S, S2).
g_data(T, X, E, S, S2) :- data(X), known(X) | r(T, yes, E, S, S2).
g_data(T, X, E, S, S2) :- data(X), unknown(X) | r(T, no, E, S, S2).
g_string(T, X, E, S, S2) :- string(X) | r(T, yes, E, S, S2).
g_string(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
g_tuple(T, X, E, S, S2) :- tuple(X) | r(T, yes, E, S, S2).
g_tuple(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
g_integer(T, X, E, S, S2) :- integer(X) | r(T, yes, E, S, S2).
g_integer(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
g_list(T, X, E, S, S2) :- list(X) | r(T, yes, E, S, S2).
g_list(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
g_module(T, X, E, S, S2) :- module(X) | r(T, yes, E, S, S2).
g_module(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
g_file_exists(T, X, E, S, S2) :- file_exists(X) | r(T, yes, E, S, S2).
g_file_exists(T, X, E, S, S2) :- otherwise | r(T, no, E, S, S2).
gto(T, X, Y, E, S, S2) :- X @> Y | r(T, yes, E, S, S2).
gto(T, X, Y, E, S, S2) :- X @=< Y | r(T, no, E, S, S2).

report({OK, 0}) :-
    writeln('all tests succeeded.').
report({OK, NOTOK}) :-
    writeln(tests_succeeded(OK)),
    writeln(tests_failed(NOTOK)).

main :- 
    S1 := {0, 0}, 
    builtins(S1, S2),
    guards(S2, S3),
    order(S3, S4),
    exprs(S4, S5),
    misc(S5, S6),
    report(S6).

builtins(S0, S) :-
    string_to_list('', L0, []), list_to_string(L0, A0),
        r(string_to_list_0, A0, '', S0, S1),
    string_to_list(abc, L1, []), list_to_string(L1, A1), 
        r(string_to_list, A1, abc, S1, S2),
    string_to_list('a⎕c', L2, []), list_to_string(L2, A2), 
        r(string_to_list_utf, A2, 'a⎕c', S2, S3),
    string_to_list('abc', L3, [0'd]), list_to_string(L3, A3), 
        r(string_to_list_tail, A3, 'abcd', S3, S4),
    string_to_byte_list('a⎕c', L2b, []), 
        r(string_to_byte_list_utf, L2b, [16'61, 16'e2, 16'8e, 16'95, 16'63], 
            S4, S4b),
    make_tuple(3, T1), length(T1, N3), 
        r(make_tuple, N3, 3, S4b, S5),
    integer_to_list(123, L4), list_to_integer(L4, N4), 
        r(integer_to_list, N4, 123, S5, S6),
    integer_to_list(256, 16, L4x), list_to_integer(L4x, 8, N4x), 
        r(integer_to_list2, N4x, 64, S6, S6x), 
    tuple_to_list({a, b, c}, L5, [d]), list_to_tuple(L5, T5), 
        r(tuple_to_list, T5, {a, b, c, d}, S6x, S7),
    tuple_to_list({}, L5x, []), list_to_tuple(L5x, T5x), 
        r(tuple_to_list, T5x, {}, S7, S7x),
    length({a, b, c}, N6), r(tuple_length, N6, 3, S7x, S8),
    length([a, b, c], N7), r(list_length, N7, 3, S8, S9),
    length(abc, N8), r(string_length, N8, 3, S9, S10),
    length([], N9), r(null_length, N9, 0, S10, S11),
    get_module(builtins, M1), run(M1, e(123, N10)), 
        r(run, N10, 246, S11, S12x),
    run(builtins, e(100, N10x)), r(run2, N10x, 200, S12x, S12),
    assign(N12, 123, L11), r(assign, {N12, L11}, {123, []}, S12, S13),
    make_tuple(3, T13), put_arg(1, T13, abc), put_arg(2, T13, 123),
        put_arg(3, T13, 456, L14), 
        r(put_arg, {T13, L14}, {abc(123, 456), []}, S13, S14),
    get_arg(1, T13, A15), get_arg(2, T13, N16), 
        get_arg(3, T13, N17, L18),
        r(get_arg, {A15, N16, N17, L18}, {abc, 123, 456, []}, S14, S15),
    open_file('test.out', w, F1),
    write_chars(F1, [65, 66, 16'2501, 67]),
    write(F1, 'xyz'),
    close_file(F1),
    MT1 is time,
    open_file('test.out', r, F2),
    read_bytes(F2, 100, FL1, []), list_to_string(FL1, FA1),
        r(file, FA1, 'ABâCxyz', S15, S16),
    read_bytes(F2, 100, FL2, _), r(file2, FL2, [], S16, S17),
    utf_decode([65, 66], U1, UL1), r(utf_decode1, U1, 65, S17, S18),
        r(utf_decode2, UL1, [66], S18, S19),
    utf_decode([16'ce, 16'9e, 16'0a], U2, UL2), 
        r(utf_decode3, U2, 926, S19, S20),
        r(utf_decode4, UL2, [10], S20, S21),
    open_pipe(PIN, POUT),
    pipe_test(PIN, POUT, PR),
        r(pipe, PR, [65, 66, 67], S21, S22),
    shell('exit 123', XS), r(shell, XS, 123, S22, S23),
    file_size('test.out', FS), r(file_size, FS, 9, S23, S24),
    file_modification_time('test.out', MT2), r(file_mtime, MT1, MT2, S24, S25),
    listen_test('test.out', LOK), r(listen, LOK, ok, S25, S26),
    string_to_integer('123', SI1), r(string_to_int, SI1, 123, S26, S27),
    string_to_integer('-123', SI2), r(string_to_int, SI2, -123, S27, S28),
    string_to_integer('1011', 2, SI3), r(string_to_int, SI3, 11, S28, S29),
    S := S29.

guards(S1, S) :-    
    eq(eq1, 123, 123, yes, S1, S2),
    eq(eq2, 123, 124, no, S2, S3),
    eq(eq3, abc, abc, yes, S3, S4),
    eq(eq4, abc, def, no, S4, S5),
    eq(eq5, [] , [], yes, S5, S6), 
    eq(eq6, {a, b} , {a, b}, yes, S6, S7),
    eq(eq7, {a, b} , {a, c}, no, S7, S8), S9 := S8,
    eq(eq8, [a, b] , [a, b], yes, S9, S10),
    eq(eq9, [a, b] , [a, c], no, S10, S11),
    eq(eq10, [a, b] , {a, b}, no, S11, S12),
    eq(eq11, [a, b] , [a], no, S12, S13),
    eq(eq11, [a, b] , [a, b, c], no, S13, S14),
    eq(eq12, {a, b} , {a}, no, S14, S15),
    eq(eq13, {a, b} , {a, b, c}, no, S15, S16),
    gt(gt1, 124, 123, yes, S16, S17),
    gt(gt2, 123, 124, no, S17, S18),
    gt(gt3, 123, 123, no, S18, S19), S20 := S19,
    lt(lt1, 123, 124, yes, S20, S21),
    lt(lt2, 124, 123, no, S21, S22),
    lt(lt3, 123, 123, no, S22, S23),
    ge(ge1, 124, 123, yes, S23, S24),
    ge(ge2, 123, 123, yes, S24, S25),
    ge(ge3, 123, 124, no, S25, S26),
    le(le1, 123, 124, yes, S26, S27),
    le(le2, 123, 123, yes, S27, S28),
    le(le3, 124, 123, no, S28, S29),
    g_known(known, 123, yes, S29, S30),
    g_known(known, _, no, S30, S31),
    g_data(data, 123, yes, S31, S32),
    g_string(string, abc, yes, S32, S33),
    g_string(no_string, 123, no, S33, S34),
    g_string(null_no_string, [], no, S34, S35),
    g_tuple(tuple, {123}, yes, S35, S36),
    g_tuple(no_tuple, 123, no, S36, S37),
    g_tuple(empty_tuple, {}, yes, S37, S38),
    g_tuple(no_null_tuple, [], no, S38, S39),
    g_integer(integer, 123, yes, S39, S40),
    g_integer(no_integer, abc, no, S40, S41),
    g_list(list, [1], yes, S41, S42),
    g_list(list, [], no, S42, S43),
    get_module(builtins, M),
    g_module(module, M, yes, S43, S44),
    g_module(no_module, {}, no, S44, S45),
    g_file_exists(exists, 'builtins.sm', yes, S45, S46),
    g_file_exists(exists_not, 'abcxyz123', no, S46, S47),
    S := S47.

order(S1, S) :-
    gto(gto1, 124, 123, yes, S1, S2),
    gto(gto1, 123, 123, no, S2, S3),
    gto(gto3, 123, 124, no, S3, S4),
    gto(gto4, abc, abd, no, S4, S5),
    gto(gto5, abd, abc, yes, S5, S6),
    gto(gto6, abc, abc, no, S6, S7),
    gto(gto7, abcd, abc, yes, S7, S8),
    gto(gto8, abc, abcd, no, S8, S9),
    gto(gto9, {a, c}, {a, b}, yes, S9, S10),
    gto(gto10, {a, b}, {a, b}, no, S10, S11),
    gto(gto11, {a, b}, {a, c}, no, S11, S12),
    gto(gto12, {a, b, c}, {a, b}, yes, S12, S13),
    gto(gto13, {a, b}, {a, b, c}, no, S13, S14),
    gto(gto14, [a|c], [a|b], yes, S14, S15),
    gto(gto15, [a|b], [a|b], no, S15, S16),
    gto(gto16, [a|b], [a|c], no, S16, S17),
    gto(gto17, {a}, [a], yes, S17, S18),
    gto(gto18, [a], a, yes, S18, S19),
    gto(gto19, a, 123, yes, S19, S20),
    gto(gto20, [a], {a}, no, S20, S21),
    gto(gto21, a, [a], no, S21, S22),
    gto(gto22, 123, a, no, S22, S23),
    gto(gto23, {}, [], yes, S23, S24),
    S := S24.

exprs(S1, S) :-
    R1 is 123 + 1, r(add1, R1, 124, S1, S2),
    R2 is 123 + (-1), r(add2, R2, 122, S2, S3),
    R3 is 123 - 1, r(sub1, R3, 122, S3, S4),
    R4 is 123 - (-1), r(sub2, R4, 124, S4, S5),
    R5 is 123 * 2, r(mul1, R5, 246, S5, S6),
    R6 is 123 * -2, r(mul2, R6, -246, S6, S7),
    R7 is 123 * 1, r(mul3, R7, 123, S7, S8),
    R8 is 123 * 0, r(mul4, R8, 0, S8, S9),
    R9 is 120 / 2, r(div1, R9, 60, S9, S10),
    R10 is 120 / 1, r(div2, R10, 120, S10, S11),
    R11 is 0 / 2, r(div3, R11, 0, S11, S12),
    R12 is 123 \\ 2, r(mod1, R12, 1, S12, S13),
    R12a is -123 \\ 2, r(mod2, R12a, -1, S13, S14),
    R13 is 123 \\ -2, r(mod3, R13, 1, S14, S15),
    R14 is abs(123), r(abs1, R14, 123, S15, S16),
    R15 is abs(-123), r(abs2, R15, 123, S16, S17),
    R16 is abs(0), r(abs3, R16, 0, S17, S18),
    R17 is 2'1101 /\ 2'1001, r(and, R17, 2'1001, S18, S19),
    R18 is 2'1101 \/ 2'1001, r(or, R18, 2'1101, S19, S20),
    R19 is 2'1101 >< 2'1001, r(xor, R19, 2'0100, S20, S21),
    R20 is \2'1101, r(not, R20, -14, S21, S22),
    R21 is 2'11011 >> 1, r(rshift, R21, 2'1101, S22, S23),
    R22 is 2'11011 << 2, r(lshift, R22, 2'1101100, S23, S24),
    N1 := 123,
    R23 is -N1, r(neg, R23, -123, S24, S25),
    S := S25.

misc(S1, S) :-
    builtins:e(100, X), r(call, X, 200, S1, S2),
    merge1(S2, S3),
    merge2(S3, S4),
    S := S4.

merge1(S1, S) :-
    merger(I, O),
    sum(0, 10, 1, I),
    mresult(O, 0, R), r(merge1, R, 55, S1, S).

merge2(S1, S) :-
    merger([merge(I1), merge(I2)], O),
    sum(0, 10, 1, I1),
    sum(100, 150, 5, I2),
    mresult(O, 0, R), r(merge2, R, 1330, S1, S).

sum(START, END, _, S) :- START >= END | S := [].
sum(START, END, STEP, S) :- 
    otherwise | 
    N is START + STEP,
    S := [N|S2],
    sum(N, END, STEP, S2).

mresult([], T, R) :- R := T.
mresult([X|S], T, R) :- T2 is T + X, mresult(S, T2, R).

pipe_test(IN, OUT, R) :-
    write_bytes(OUT, [65, 66, 67], OK),
    pipe_test1(OK, IN, OUT, R).

pipe_test1([], IN, OUT, R) :-
    close_file(OUT),
    read_bytes(IN, 3, R, []),
    done_close(R, IN).

done_close(X, F) :-
    data(X) | close_file(F).

listen_test(FN, OK) :-
    open_file(FN, r, FD),
    listen(FD, INP),
    THEN is time + 5,
    listen_wait(0, THEN, FD, INP, OK).

listen_wait(_, _, FD, INP, OK) :-
    known(INP) | 
    close_file(FD), OK := ok.
listen_wait(1000, THEN, FD, INP, OK) :-
    unknown(INP) |
    NOW is time,
    listen_wait2(NOW, THEN, FD, INP, OK).
listen_wait(T, THEN, FD, INP, OK) :-
    unknown(INP) |
    T2 is T + 1,
    listen_wait(T2, THEN, FD, INP, OK).
listen_wait(NOW, THEN, FD, INP, OK) :-
    otherwise |
    % just repeat - wakeup due to listening may be triggered
    % in-between matching attempts...
    listen_wait(NOW, THEN, FD, INP, OK).

listen_wait2(NOW, THEN, FD, INP, OK) :-
    NOW >= THEN |
    close_file(FD),
    OK := timeout.
listen_wait2(NOW, THEN, FD, INP, OK) :-
    NOW < THEN |
    listen_wait(0, THEN, FD, INP, OK).
